/***********************************************************************
 * File: byteorder.c
 * 	(Extracted from
 *	   "$Id: byteorder.c,v 1.4 2007/10/21 06:16:04 Administrator Exp Administrator $")
 *
 * Abstracts:
 *  This C source file implements `byteorder' API-s.
 *
 * Copyright (C) 2007 Maoyam Tokyo, Japan (mailto:maoyam@mail.goo.ne.jp)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 **********************************************************************/

#include "byteorder.h"
#include <arpa/inet.h>


/* Macro: transfer integers  [ */
#define Conv_integers(TYPE,NTOH,HTON) {				\
	TYPE	val, new_val;					\
	while (occurrence-- > 0) {				\
		memcpy(&val, src, sizeof(val));			\
		new_val	= ((direction == Byteorder_dir_e_ntoh)?	\
					NTOH(val): HTON(val)	\
				);				\
		memcpy(dst, &new_val, sizeof(new_val));		\
		src += sizeof(val); dst += sizeof(val);		\
	}							\
}

/* transfer unsigned 32bits integer sequence */
static void conv_longs(Byteorder_dir_e	direction
			, char		*src
			, char		*dst
			, size_t	occurrence
			)
{
	Conv_integers(uint32_t, ntohl, htonl);
	return;
}

/* transfer unsigned 16bits integer sequence */
static void conv_shorts(Byteorder_dir_e	direction
			, char		*src
			, char		*dst
			, size_t	occurrence
			)
{
	Conv_integers(uint16_t, ntohs, htons);
	return;
}

#undef Conv_integers /* ] Macro: transfer integers */

/* transfer the octets (unsigned 8bits integer) sequence */
static void conv_chars(char		*src
			, char		*dst
			, size_t	occurrence
			)
{
	/* test: `src' is not the same address as `dst'? */
	if (src != dst) {
		/* test: `dst' buffer doesn't overlap `src'? */
		if (dst < src || (src + occurrence) <= dst) {
			/* transfer buffers (as is) */
			memcpy(dst, src, occurrence);
		}
		else {	/* transfer overlapped buffers */
			for (src += (occurrence - 1)
				, dst += (occurrence - 1)
			; occurrence-- > 0; --src, --dst) {
				*dst	= *src;
			}
		}
	}

	return;
}


/* Byteorder converting for the sequence of 32bits integers
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the sequence of network-byteorder 32bits
 *	 integers.
 *	o `hostp' points to the sequence of host-byteorder 32bits
 *	 integers.
 *	o `occurrence' indicates the number of integers sequences.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder_longs(Byteorder_dir_e	direction
			, void			*netp
			, void			*hostp
			, size_t		occurrence
			)
{
	char	*net = (char *)netp, *host = (char *)hostp;

	/* test: invalid arguments? */
	if (((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| net == NULL) || host == NULL) || occurrence == 0) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}

	if (direction == Byteorder_dir_e_ntoh) {
		/* convert it (network-byteorder to host's one) */
		conv_longs(direction, net, host, occurrence);
		return	host; /* return normally */
	}
	else {	/* convert it (host-byteordr to network's one) */
		conv_longs(direction, host, net, occurrence);
		return	net; /* return normally */
	}
}

/* Byteorder converting for the sequence of 16bits integers
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the sequence of network-byteorder 16bits
 *	 integers.
 *	o `hostp' points to the sequence of host-byteorder 16bits
 *	 integers.
 *	o `occurrence' indicates the number of integers sequences.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder_shorts(Byteorder_dir_e	direction
			, void			*netp
			, void			*hostp
			, size_t		occurrence
			)
{
	char	*net = (char *)netp, *host = (char *)hostp;

	/* test: invalid arguments? */
	if (((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| net == NULL) || host == NULL) || occurrence == 0) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}

	if (direction == Byteorder_dir_e_ntoh) {
		/* convert it (network-byteorder to host's one) */
		conv_shorts(direction, net, host, occurrence);
		return	host; /* return normally */
	}
	else {	/* convert it (host-byteorder to network's one) */
		conv_shorts(direction, host, net, occurrence);
		return	net; /* return normally */
	}
}

/* Byteorder converting for the sequence of 8bits integers
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the sequence of network-byteorder 8bits
 *	 integers.
 *	o `hostp' points to the sequence of host-byteorder 8bits
 *	 integers.
 *	o `occurrence' indicates the number of integers sequences.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder_chars(Byteorder_dir_e	direction
			, char			*netp
			, char			*hostp
			, size_t		occurrence
			)
{
	/* test: invalid arguments? */
	if (((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| netp == NULL) || hostp == NULL) || occurrence == 0) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}

	if (direction == Byteorder_dir_e_ntoh) {
		/* convert it (network-byteorder to host's one) */
		conv_chars(netp, hostp, occurrence);
		return	hostp; /* return normally */
	}
	else {	/* convert it (host-byteorder to network's one) */
		conv_chars(hostp, netp, occurrence);
		return	netp; /* return normally */
	}
}

/* Byteorder converting for the differ buffers.
 * Arguments;
 *	o `direction' indicates converting directions;
 *		(1) Byteorder_dir_e_ntoh: network-wide to host-native
 *		(2) Byteorder_dir_e_hton: host-native to network-wide
 *	o `netp' points to the network-byteorder buffer.
 *	o `net_rec_sz' indicates the size of `netp' records.
 *	o `layouts_n' indicates the number of elements of `layouts'
 *	 array.
 *	o `layouts' points to the array of layout of the converting
 *	 fields.
 *	o `hostp' points to the host-byteorder buffer.
 *	o `host_rec_sz' indicates the size of `hostp' records.
 * Returns;
 *	NULL    : several errors occurred.
 *	non-NULL: return normally
 */
char	*byteorder(Byteorder_dir_e	direction
		, void			*netp
		, size_t		net_rec_sz
		, size_t		layouts_n
		, Byteorder_layout_t	*layouts
		, void			*hostp
		, size_t		host_rec_sz
		)
{
	size_t	i;
	char	*net = (char *)netp, *host = (char *)hostp;

	/* test: invalid arguments? */
	if ((((((((direction < Byteorder_dir_e_ntoh)
	|| direction >= Byteorder_dir_e_N)
	|| net == NULL)		|| net_rec_sz == 0)
	|| layouts_n == 0)	|| layouts == NULL)
	|| host == NULL)	|| host_rec_sz == 0) {
		/* ERROR: invalid arguments */
		errno = EINVAL; return NULL; /* return error */
	}
	for (i = 0; i < layouts_n; ++i) {
		if (((((layouts[i].type < Byteorder_type_e_char)
		|| layouts[i].type >= Byteorder_type_e_N)
		|| layouts[i].occurrence == 0)
		|| (layouts[i].net_fld_offset
			+ (Byteorder_type_to_size(layouts[i].type)
				* layouts[i].occurrence)) > net_rec_sz)
		|| (layouts[i].host_fld_offset
			+ (Byteorder_type_to_size(layouts[i].type)
				* layouts[i].occurrence)) > host_rec_sz)
		{
			/* ERROR: invalid `layouts' argument */
			errno = EINVAL; return NULL; /* return error */
	}	}

	/* transfer the each fields... */
	for (i = 0; i < layouts_n; ++layouts, ++i) {
		switch (layouts->type) {
		case Byteorder_type_e_long: /* unsigned 32bits */
			if (direction == Byteorder_dir_e_ntoh) {
				conv_longs(direction
				, (net + layouts->net_fld_offset)
				, (host + layouts->host_fld_offset)
				, (layouts->occurrence)
				);
			}
			else {
				conv_longs(direction
				, (host + layouts->host_fld_offset)
				, (net + layouts->net_fld_offset)
				, (layouts->occurrence)
				);
			}
			break;

		case Byteorder_type_e_short: /* unsigned 16bits */
			if (direction == Byteorder_dir_e_ntoh) {
				conv_shorts(direction
				, (net + layouts->net_fld_offset)
				, (host + layouts->host_fld_offset)
				, (layouts->occurrence)
				);
			}
			else {
				conv_shorts(direction
				, (host + layouts->host_fld_offset)
				, (net + layouts->net_fld_offset)
				, (layouts->occurrence)
				);
			}
			break;

		case Byteorder_type_e_char: /* octets (8bits) */
		default:
			if (direction == Byteorder_dir_e_ntoh) {
				conv_chars(
				  (net + layouts->net_fld_offset)
				, (host + layouts->host_fld_offset)
				, (layouts->occurrence)
				);
			}
			else {
				conv_chars(
				  (host + layouts->host_fld_offset)
				, (net + layouts->net_fld_offset)
				, (layouts->occurrence)
				);
			}
			break;
		}
	}

	/* return normally */
	return	(direction == Byteorder_dir_e_ntoh)? host: net;
}
